# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: wandb/proto/wandb_server.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from wandb.proto import wandb_base_pb2 as wandb_dot_proto_dot_wandb__base__pb2
from wandb.proto import wandb_internal_pb2 as wandb_dot_proto_dot_wandb__internal__pb2
from wandb.proto import wandb_telemetry_pb2 as wandb_dot_proto_dot_wandb__telemetry__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='wandb/proto/wandb_server.proto',
  package='wandb_internal',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1ewandb/proto/wandb_server.proto\x12\x0ewandb_internal\x1a\x1cwandb/proto/wandb_base.proto\x1a wandb/proto/wandb_internal.proto\x1a!wandb/proto/wandb_telemetry.proto\"D\n\x15ServerShutdownRequest\x12+\n\x05_info\x18\xc8\x01 \x01(\x0b\x32\x1b.wandb_internal._RecordInfo\"\x18\n\x16ServerShutdownResponse\"B\n\x13ServerStatusRequest\x12+\n\x05_info\x18\xc8\x01 \x01(\x0b\x32\x1b.wandb_internal._RecordInfo\"\x16\n\x14ServerStatusResponse\")\n\x10StringTupleValue\x12\x15\n\rstring_values\x18\x01 \x03(\t\"\xe1\x01\n\rSettingsValue\x12\x13\n\tint_value\x18\x01 \x01(\x03H\x00\x12\x16\n\x0cstring_value\x18\x02 \x01(\tH\x00\x12\x15\n\x0b\x66loat_value\x18\x03 \x01(\x01H\x00\x12\x14\n\nbool_value\x18\x04 \x01(\x08H\x00\x12\x14\n\nnull_value\x18\x05 \x01(\x08H\x00\x12\x37\n\x0btuple_value\x18\x06 \x01(\x0b\x32 .wandb_internal.StringTupleValueH\x00\x12\x19\n\x0ftimestamp_value\x18\x07 \x01(\tH\x00\x42\x0c\n\nvalue_type\"\xea\x01\n\x17ServerInformInitRequest\x12O\n\r_settings_map\x18\x32 \x03(\x0b\x32\x38.wandb_internal.ServerInformInitRequest.SettingsMapEntry\x12+\n\x05_info\x18\xc8\x01 \x01(\x0b\x32\x1b.wandb_internal._RecordInfo\x1aQ\n\x10SettingsMapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x1d.wandb_internal.SettingsValue:\x02\x38\x01\"\x1a\n\x18ServerInformInitResponse\"\xec\x01\n\x18ServerInformStartRequest\x12P\n\r_settings_map\x18\x32 \x03(\x0b\x32\x39.wandb_internal.ServerInformStartRequest.SettingsMapEntry\x12+\n\x05_info\x18\xc8\x01 \x01(\x0b\x32\x1b.wandb_internal._RecordInfo\x1aQ\n\x10SettingsMapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x1d.wandb_internal.SettingsValue:\x02\x38\x01\"\x1b\n\x19ServerInformStartResponse\"H\n\x19ServerInformFinishRequest\x12+\n\x05_info\x18\xc8\x01 \x01(\x0b\x32\x1b.wandb_internal._RecordInfo\"\x1c\n\x1aServerInformFinishResponse\"H\n\x19ServerInformAttachRequest\x12+\n\x05_info\x18\xc8\x01 \x01(\x0b\x32\x1b.wandb_internal._RecordInfo\"\xf0\x01\n\x1aServerInformAttachResponse\x12R\n\r_settings_map\x18\x32 \x03(\x0b\x32;.wandb_internal.ServerInformAttachResponse.SettingsMapEntry\x12+\n\x05_info\x18\xc8\x01 \x01(\x0b\x32\x1b.wandb_internal._RecordInfo\x1aQ\n\x10SettingsMapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12,\n\x05value\x18\x02 \x01(\x0b\x32\x1d.wandb_internal.SettingsValue:\x02\x38\x01\"H\n\x19ServerInformDetachRequest\x12+\n\x05_info\x18\xc8\x01 \x01(\x0b\x32\x1b.wandb_internal._RecordInfo\"\x1c\n\x1aServerInformDetachResponse\"]\n\x1bServerInformTeardownRequest\x12\x11\n\texit_code\x18\x01 \x01(\x05\x12+\n\x05_info\x18\xc8\x01 \x01(\x0b\x32\x1b.wandb_internal._RecordInfo\"\x1e\n\x1cServerInformTeardownResponse\"\xa4\x04\n\rServerRequest\x12\x30\n\x0erecord_publish\x18\x01 \x01(\x0b\x32\x16.wandb_internal.RecordH\x00\x12\x34\n\x12record_communicate\x18\x02 \x01(\x0b\x32\x16.wandb_internal.RecordH\x00\x12>\n\x0binform_init\x18\x03 \x01(\x0b\x32\'.wandb_internal.ServerInformInitRequestH\x00\x12\x42\n\rinform_finish\x18\x04 \x01(\x0b\x32).wandb_internal.ServerInformFinishRequestH\x00\x12\x42\n\rinform_attach\x18\x05 \x01(\x0b\x32).wandb_internal.ServerInformAttachRequestH\x00\x12\x42\n\rinform_detach\x18\x06 \x01(\x0b\x32).wandb_internal.ServerInformDetachRequestH\x00\x12\x46\n\x0finform_teardown\x18\x07 \x01(\x0b\x32+.wandb_internal.ServerInformTeardownRequestH\x00\x12@\n\x0cinform_start\x18\x08 \x01(\x0b\x32(.wandb_internal.ServerInformStartRequestH\x00\x42\x15\n\x13server_request_type\"\xb0\x04\n\x0eServerResponse\x12\x34\n\x12result_communicate\x18\x02 \x01(\x0b\x32\x16.wandb_internal.ResultH\x00\x12H\n\x14inform_init_response\x18\x03 \x01(\x0b\x32(.wandb_internal.ServerInformInitResponseH\x00\x12L\n\x16inform_finish_response\x18\x04 \x01(\x0b\x32*.wandb_internal.ServerInformFinishResponseH\x00\x12L\n\x16inform_attach_response\x18\x05 \x01(\x0b\x32*.wandb_internal.ServerInformAttachResponseH\x00\x12L\n\x16inform_detach_response\x18\x06 \x01(\x0b\x32*.wandb_internal.ServerInformDetachResponseH\x00\x12P\n\x18inform_teardown_response\x18\x07 \x01(\x0b\x32,.wandb_internal.ServerInformTeardownResponseH\x00\x12J\n\x15inform_start_response\x18\x08 \x01(\x0b\x32).wandb_internal.ServerInformStartResponseH\x00\x42\x16\n\x14server_response_type2\xd4\x17\n\x0fInternalService\x12I\n\tRunUpdate\x12\x19.wandb_internal.RunRecord\x1a\x1f.wandb_internal.RunUpdateResult\"\x00\x12I\n\x06\x41ttach\x12\x1d.wandb_internal.AttachRequest\x1a\x1e.wandb_internal.AttachResponse\"\x00\x12>\n\x06TBSend\x12\x18.wandb_internal.TBRecord\x1a\x18.wandb_internal.TBResult\"\x00\x12O\n\x08RunStart\x12\x1f.wandb_internal.RunStartRequest\x1a .wandb_internal.RunStartResponse\"\x00\x12U\n\nGetSummary\x12!.wandb_internal.GetSummaryRequest\x1a\".wandb_internal.GetSummaryResponse\"\x00\x12\x61\n\x0eSampledHistory\x12%.wandb_internal.SampledHistoryRequest\x1a&.wandb_internal.SampledHistoryResponse\"\x00\x12O\n\x08PollExit\x12\x1f.wandb_internal.PollExitRequest\x1a .wandb_internal.PollExitResponse\"\x00\x12O\n\x08Shutdown\x12\x1f.wandb_internal.ShutdownRequest\x1a .wandb_internal.ShutdownResponse\"\x00\x12I\n\x07RunExit\x12\x1d.wandb_internal.RunExitRecord\x1a\x1d.wandb_internal.RunExitResult\"\x00\x12[\n\rRunPreempting\x12#.wandb_internal.RunPreemptingRecord\x1a#.wandb_internal.RunPreemptingResult\"\x00\x12\x46\n\x06Metric\x12\x1c.wandb_internal.MetricRecord\x1a\x1c.wandb_internal.MetricResult\"\x00\x12]\n\nPartialLog\x12%.wandb_internal.PartialHistoryRequest\x1a&.wandb_internal.PartialHistoryResponse\"\x00\x12\x45\n\x03Log\x12\x1d.wandb_internal.HistoryRecord\x1a\x1d.wandb_internal.HistoryResult\"\x00\x12I\n\x07Summary\x12\x1d.wandb_internal.SummaryRecord\x1a\x1d.wandb_internal.SummaryResult\"\x00\x12\x46\n\x06\x43onfig\x12\x1c.wandb_internal.ConfigRecord\x1a\x1c.wandb_internal.ConfigResult\"\x00\x12\x43\n\x05\x46iles\x12\x1b.wandb_internal.FilesRecord\x1a\x1b.wandb_internal.FilesResult\"\x00\x12\x46\n\x06Output\x12\x1c.wandb_internal.OutputRecord\x1a\x1c.wandb_internal.OutputResult\"\x00\x12O\n\tTelemetry\x12\x1f.wandb_internal.TelemetryRecord\x1a\x1f.wandb_internal.TelemetryResult\"\x00\x12\x43\n\x05\x41lert\x12\x1b.wandb_internal.AlertRecord\x1a\x1b.wandb_internal.AlertResult\"\x00\x12L\n\x08\x41rtifact\x12\x1e.wandb_internal.ArtifactRecord\x1a\x1e.wandb_internal.ArtifactResult\"\x00\x12X\n\x0cLinkArtifact\x12\".wandb_internal.LinkArtifactRecord\x1a\".wandb_internal.LinkArtifactResult\"\x00\x12[\n\x0c\x41rtifactSend\x12#.wandb_internal.ArtifactSendRequest\x1a$.wandb_internal.ArtifactSendResponse\"\x00\x12[\n\x0c\x41rtifactPoll\x12#.wandb_internal.ArtifactPollRequest\x1a$.wandb_internal.ArtifactPollResponse\"\x00\x12[\n\x0c\x43heckVersion\x12#.wandb_internal.CheckVersionRequest\x1a$.wandb_internal.CheckVersionResponse\"\x00\x12\x46\n\x05Pause\x12\x1c.wandb_internal.PauseRequest\x1a\x1d.wandb_internal.PauseResponse\"\x00\x12I\n\x06Resume\x12\x1d.wandb_internal.ResumeRequest\x1a\x1e.wandb_internal.ResumeResponse\"\x00\x12I\n\x06Status\x12\x1d.wandb_internal.StatusRequest\x1a\x1e.wandb_internal.StatusResponse\"\x00\x12\x61\n\x0eServerShutdown\x12%.wandb_internal.ServerShutdownRequest\x1a&.wandb_internal.ServerShutdownResponse\"\x00\x12[\n\x0cServerStatus\x12#.wandb_internal.ServerStatusRequest\x1a$.wandb_internal.ServerStatusResponse\"\x00\x12g\n\x10ServerInformInit\x12\'.wandb_internal.ServerInformInitRequest\x1a(.wandb_internal.ServerInformInitResponse\"\x00\x12j\n\x11ServerInformStart\x12(.wandb_internal.ServerInformStartRequest\x1a).wandb_internal.ServerInformStartResponse\"\x00\x12m\n\x12ServerInformFinish\x12).wandb_internal.ServerInformFinishRequest\x1a*.wandb_internal.ServerInformFinishResponse\"\x00\x12m\n\x12ServerInformAttach\x12).wandb_internal.ServerInformAttachRequest\x1a*.wandb_internal.ServerInformAttachResponse\"\x00\x12m\n\x12ServerInformDetach\x12).wandb_internal.ServerInformDetachRequest\x1a*.wandb_internal.ServerInformDetachResponse\"\x00\x12s\n\x14ServerInformTeardown\x12+.wandb_internal.ServerInformTeardownRequest\x1a,.wandb_internal.ServerInformTeardownResponse\"\x00\x62\x06proto3'
  ,
  dependencies=[wandb_dot_proto_dot_wandb__base__pb2.DESCRIPTOR,wandb_dot_proto_dot_wandb__internal__pb2.DESCRIPTOR,wandb_dot_proto_dot_wandb__telemetry__pb2.DESCRIPTOR,])




_SERVERSHUTDOWNREQUEST = _descriptor.Descriptor(
  name='ServerShutdownRequest',
  full_name='wandb_internal.ServerShutdownRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='_info', full_name='wandb_internal.ServerShutdownRequest._info', index=0,
      number=200, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=149,
  serialized_end=217,
)


_SERVERSHUTDOWNRESPONSE = _descriptor.Descriptor(
  name='ServerShutdownResponse',
  full_name='wandb_internal.ServerShutdownResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=219,
  serialized_end=243,
)


_SERVERSTATUSREQUEST = _descriptor.Descriptor(
  name='ServerStatusRequest',
  full_name='wandb_internal.ServerStatusRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='_info', full_name='wandb_internal.ServerStatusRequest._info', index=0,
      number=200, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=245,
  serialized_end=311,
)


_SERVERSTATUSRESPONSE = _descriptor.Descriptor(
  name='ServerStatusResponse',
  full_name='wandb_internal.ServerStatusResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=313,
  serialized_end=335,
)


_STRINGTUPLEVALUE = _descriptor.Descriptor(
  name='StringTupleValue',
  full_name='wandb_internal.StringTupleValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='string_values', full_name='wandb_internal.StringTupleValue.string_values', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=337,
  serialized_end=378,
)


_SETTINGSVALUE = _descriptor.Descriptor(
  name='SettingsValue',
  full_name='wandb_internal.SettingsValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='int_value', full_name='wandb_internal.SettingsValue.int_value', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='string_value', full_name='wandb_internal.SettingsValue.string_value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='float_value', full_name='wandb_internal.SettingsValue.float_value', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bool_value', full_name='wandb_internal.SettingsValue.bool_value', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='null_value', full_name='wandb_internal.SettingsValue.null_value', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tuple_value', full_name='wandb_internal.SettingsValue.tuple_value', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp_value', full_name='wandb_internal.SettingsValue.timestamp_value', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value_type', full_name='wandb_internal.SettingsValue.value_type',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=381,
  serialized_end=606,
)


_SERVERINFORMINITREQUEST_SETTINGSMAPENTRY = _descriptor.Descriptor(
  name='SettingsMapEntry',
  full_name='wandb_internal.ServerInformInitRequest.SettingsMapEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='wandb_internal.ServerInformInitRequest.SettingsMapEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='wandb_internal.ServerInformInitRequest.SettingsMapEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=762,
  serialized_end=843,
)

_SERVERINFORMINITREQUEST = _descriptor.Descriptor(
  name='ServerInformInitRequest',
  full_name='wandb_internal.ServerInformInitRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='_settings_map', full_name='wandb_internal.ServerInformInitRequest._settings_map', index=0,
      number=50, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='_info', full_name='wandb_internal.ServerInformInitRequest._info', index=1,
      number=200, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_SERVERINFORMINITREQUEST_SETTINGSMAPENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=609,
  serialized_end=843,
)


_SERVERINFORMINITRESPONSE = _descriptor.Descriptor(
  name='ServerInformInitResponse',
  full_name='wandb_internal.ServerInformInitResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=845,
  serialized_end=871,
)


_SERVERINFORMSTARTREQUEST_SETTINGSMAPENTRY = _descriptor.Descriptor(
  name='SettingsMapEntry',
  full_name='wandb_internal.ServerInformStartRequest.SettingsMapEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='wandb_internal.ServerInformStartRequest.SettingsMapEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='wandb_internal.ServerInformStartRequest.SettingsMapEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=762,
  serialized_end=843,
)

_SERVERINFORMSTARTREQUEST = _descriptor.Descriptor(
  name='ServerInformStartRequest',
  full_name='wandb_internal.ServerInformStartRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='_settings_map', full_name='wandb_internal.ServerInformStartRequest._settings_map', index=0,
      number=50, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='_info', full_name='wandb_internal.ServerInformStartRequest._info', index=1,
      number=200, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_SERVERINFORMSTARTREQUEST_SETTINGSMAPENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=874,
  serialized_end=1110,
)


_SERVERINFORMSTARTRESPONSE = _descriptor.Descriptor(
  name='ServerInformStartResponse',
  full_name='wandb_internal.ServerInformStartResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1112,
  serialized_end=1139,
)


_SERVERINFORMFINISHREQUEST = _descriptor.Descriptor(
  name='ServerInformFinishRequest',
  full_name='wandb_internal.ServerInformFinishRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='_info', full_name='wandb_internal.ServerInformFinishRequest._info', index=0,
      number=200, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1141,
  serialized_end=1213,
)


_SERVERINFORMFINISHRESPONSE = _descriptor.Descriptor(
  name='ServerInformFinishResponse',
  full_name='wandb_internal.ServerInformFinishResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1215,
  serialized_end=1243,
)


_SERVERINFORMATTACHREQUEST = _descriptor.Descriptor(
  name='ServerInformAttachRequest',
  full_name='wandb_internal.ServerInformAttachRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='_info', full_name='wandb_internal.ServerInformAttachRequest._info', index=0,
      number=200, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1245,
  serialized_end=1317,
)


_SERVERINFORMATTACHRESPONSE_SETTINGSMAPENTRY = _descriptor.Descriptor(
  name='SettingsMapEntry',
  full_name='wandb_internal.ServerInformAttachResponse.SettingsMapEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='wandb_internal.ServerInformAttachResponse.SettingsMapEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='wandb_internal.ServerInformAttachResponse.SettingsMapEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=762,
  serialized_end=843,
)

_SERVERINFORMATTACHRESPONSE = _descriptor.Descriptor(
  name='ServerInformAttachResponse',
  full_name='wandb_internal.ServerInformAttachResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='_settings_map', full_name='wandb_internal.ServerInformAttachResponse._settings_map', index=0,
      number=50, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='_info', full_name='wandb_internal.ServerInformAttachResponse._info', index=1,
      number=200, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_SERVERINFORMATTACHRESPONSE_SETTINGSMAPENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1320,
  serialized_end=1560,
)


_SERVERINFORMDETACHREQUEST = _descriptor.Descriptor(
  name='ServerInformDetachRequest',
  full_name='wandb_internal.ServerInformDetachRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='_info', full_name='wandb_internal.ServerInformDetachRequest._info', index=0,
      number=200, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1562,
  serialized_end=1634,
)


_SERVERINFORMDETACHRESPONSE = _descriptor.Descriptor(
  name='ServerInformDetachResponse',
  full_name='wandb_internal.ServerInformDetachResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1636,
  serialized_end=1664,
)


_SERVERINFORMTEARDOWNREQUEST = _descriptor.Descriptor(
  name='ServerInformTeardownRequest',
  full_name='wandb_internal.ServerInformTeardownRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='exit_code', full_name='wandb_internal.ServerInformTeardownRequest.exit_code', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='_info', full_name='wandb_internal.ServerInformTeardownRequest._info', index=1,
      number=200, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1666,
  serialized_end=1759,
)


_SERVERINFORMTEARDOWNRESPONSE = _descriptor.Descriptor(
  name='ServerInformTeardownResponse',
  full_name='wandb_internal.ServerInformTeardownResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1761,
  serialized_end=1791,
)


_SERVERREQUEST = _descriptor.Descriptor(
  name='ServerRequest',
  full_name='wandb_internal.ServerRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='record_publish', full_name='wandb_internal.ServerRequest.record_publish', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='record_communicate', full_name='wandb_internal.ServerRequest.record_communicate', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inform_init', full_name='wandb_internal.ServerRequest.inform_init', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inform_finish', full_name='wandb_internal.ServerRequest.inform_finish', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inform_attach', full_name='wandb_internal.ServerRequest.inform_attach', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inform_detach', full_name='wandb_internal.ServerRequest.inform_detach', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inform_teardown', full_name='wandb_internal.ServerRequest.inform_teardown', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inform_start', full_name='wandb_internal.ServerRequest.inform_start', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='server_request_type', full_name='wandb_internal.ServerRequest.server_request_type',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1794,
  serialized_end=2342,
)


_SERVERRESPONSE = _descriptor.Descriptor(
  name='ServerResponse',
  full_name='wandb_internal.ServerResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='result_communicate', full_name='wandb_internal.ServerResponse.result_communicate', index=0,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inform_init_response', full_name='wandb_internal.ServerResponse.inform_init_response', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inform_finish_response', full_name='wandb_internal.ServerResponse.inform_finish_response', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inform_attach_response', full_name='wandb_internal.ServerResponse.inform_attach_response', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inform_detach_response', full_name='wandb_internal.ServerResponse.inform_detach_response', index=4,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inform_teardown_response', full_name='wandb_internal.ServerResponse.inform_teardown_response', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='inform_start_response', full_name='wandb_internal.ServerResponse.inform_start_response', index=6,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='server_response_type', full_name='wandb_internal.ServerResponse.server_response_type',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=2345,
  serialized_end=2905,
)

_SERVERSHUTDOWNREQUEST.fields_by_name['_info'].message_type = wandb_dot_proto_dot_wandb__base__pb2.__RECORDINFO
_SERVERSTATUSREQUEST.fields_by_name['_info'].message_type = wandb_dot_proto_dot_wandb__base__pb2.__RECORDINFO
_SETTINGSVALUE.fields_by_name['tuple_value'].message_type = _STRINGTUPLEVALUE
_SETTINGSVALUE.oneofs_by_name['value_type'].fields.append(
  _SETTINGSVALUE.fields_by_name['int_value'])
_SETTINGSVALUE.fields_by_name['int_value'].containing_oneof = _SETTINGSVALUE.oneofs_by_name['value_type']
_SETTINGSVALUE.oneofs_by_name['value_type'].fields.append(
  _SETTINGSVALUE.fields_by_name['string_value'])
_SETTINGSVALUE.fields_by_name['string_value'].containing_oneof = _SETTINGSVALUE.oneofs_by_name['value_type']
_SETTINGSVALUE.oneofs_by_name['value_type'].fields.append(
  _SETTINGSVALUE.fields_by_name['float_value'])
_SETTINGSVALUE.fields_by_name['float_value'].containing_oneof = _SETTINGSVALUE.oneofs_by_name['value_type']
_SETTINGSVALUE.oneofs_by_name['value_type'].fields.append(
  _SETTINGSVALUE.fields_by_name['bool_value'])
_SETTINGSVALUE.fields_by_name['bool_value'].containing_oneof = _SETTINGSVALUE.oneofs_by_name['value_type']
_SETTINGSVALUE.oneofs_by_name['value_type'].fields.append(
  _SETTINGSVALUE.fields_by_name['null_value'])
_SETTINGSVALUE.fields_by_name['null_value'].containing_oneof = _SETTINGSVALUE.oneofs_by_name['value_type']
_SETTINGSVALUE.oneofs_by_name['value_type'].fields.append(
  _SETTINGSVALUE.fields_by_name['tuple_value'])
_SETTINGSVALUE.fields_by_name['tuple_value'].containing_oneof = _SETTINGSVALUE.oneofs_by_name['value_type']
_SETTINGSVALUE.oneofs_by_name['value_type'].fields.append(
  _SETTINGSVALUE.fields_by_name['timestamp_value'])
_SETTINGSVALUE.fields_by_name['timestamp_value'].containing_oneof = _SETTINGSVALUE.oneofs_by_name['value_type']
_SERVERINFORMINITREQUEST_SETTINGSMAPENTRY.fields_by_name['value'].message_type = _SETTINGSVALUE
_SERVERINFORMINITREQUEST_SETTINGSMAPENTRY.containing_type = _SERVERINFORMINITREQUEST
_SERVERINFORMINITREQUEST.fields_by_name['_settings_map'].message_type = _SERVERINFORMINITREQUEST_SETTINGSMAPENTRY
_SERVERINFORMINITREQUEST.fields_by_name['_info'].message_type = wandb_dot_proto_dot_wandb__base__pb2.__RECORDINFO
_SERVERINFORMSTARTREQUEST_SETTINGSMAPENTRY.fields_by_name['value'].message_type = _SETTINGSVALUE
_SERVERINFORMSTARTREQUEST_SETTINGSMAPENTRY.containing_type = _SERVERINFORMSTARTREQUEST
_SERVERINFORMSTARTREQUEST.fields_by_name['_settings_map'].message_type = _SERVERINFORMSTARTREQUEST_SETTINGSMAPENTRY
_SERVERINFORMSTARTREQUEST.fields_by_name['_info'].message_type = wandb_dot_proto_dot_wandb__base__pb2.__RECORDINFO
_SERVERINFORMFINISHREQUEST.fields_by_name['_info'].message_type = wandb_dot_proto_dot_wandb__base__pb2.__RECORDINFO
_SERVERINFORMATTACHREQUEST.fields_by_name['_info'].message_type = wandb_dot_proto_dot_wandb__base__pb2.__RECORDINFO
_SERVERINFORMATTACHRESPONSE_SETTINGSMAPENTRY.fields_by_name['value'].message_type = _SETTINGSVALUE
_SERVERINFORMATTACHRESPONSE_SETTINGSMAPENTRY.containing_type = _SERVERINFORMATTACHRESPONSE
_SERVERINFORMATTACHRESPONSE.fields_by_name['_settings_map'].message_type = _SERVERINFORMATTACHRESPONSE_SETTINGSMAPENTRY
_SERVERINFORMATTACHRESPONSE.fields_by_name['_info'].message_type = wandb_dot_proto_dot_wandb__base__pb2.__RECORDINFO
_SERVERINFORMDETACHREQUEST.fields_by_name['_info'].message_type = wandb_dot_proto_dot_wandb__base__pb2.__RECORDINFO
_SERVERINFORMTEARDOWNREQUEST.fields_by_name['_info'].message_type = wandb_dot_proto_dot_wandb__base__pb2.__RECORDINFO
_SERVERREQUEST.fields_by_name['record_publish'].message_type = wandb_dot_proto_dot_wandb__internal__pb2._RECORD
_SERVERREQUEST.fields_by_name['record_communicate'].message_type = wandb_dot_proto_dot_wandb__internal__pb2._RECORD
_SERVERREQUEST.fields_by_name['inform_init'].message_type = _SERVERINFORMINITREQUEST
_SERVERREQUEST.fields_by_name['inform_finish'].message_type = _SERVERINFORMFINISHREQUEST
_SERVERREQUEST.fields_by_name['inform_attach'].message_type = _SERVERINFORMATTACHREQUEST
_SERVERREQUEST.fields_by_name['inform_detach'].message_type = _SERVERINFORMDETACHREQUEST
_SERVERREQUEST.fields_by_name['inform_teardown'].message_type = _SERVERINFORMTEARDOWNREQUEST
_SERVERREQUEST.fields_by_name['inform_start'].message_type = _SERVERINFORMSTARTREQUEST
_SERVERREQUEST.oneofs_by_name['server_request_type'].fields.append(
  _SERVERREQUEST.fields_by_name['record_publish'])
_SERVERREQUEST.fields_by_name['record_publish'].containing_oneof = _SERVERREQUEST.oneofs_by_name['server_request_type']
_SERVERREQUEST.oneofs_by_name['server_request_type'].fields.append(
  _SERVERREQUEST.fields_by_name['record_communicate'])
_SERVERREQUEST.fields_by_name['record_communicate'].containing_oneof = _SERVERREQUEST.oneofs_by_name['server_request_type']
_SERVERREQUEST.oneofs_by_name['server_request_type'].fields.append(
  _SERVERREQUEST.fields_by_name['inform_init'])
_SERVERREQUEST.fields_by_name['inform_init'].containing_oneof = _SERVERREQUEST.oneofs_by_name['server_request_type']
_SERVERREQUEST.oneofs_by_name['server_request_type'].fields.append(
  _SERVERREQUEST.fields_by_name['inform_finish'])
_SERVERREQUEST.fields_by_name['inform_finish'].containing_oneof = _SERVERREQUEST.oneofs_by_name['server_request_type']
_SERVERREQUEST.oneofs_by_name['server_request_type'].fields.append(
  _SERVERREQUEST.fields_by_name['inform_attach'])
_SERVERREQUEST.fields_by_name['inform_attach'].containing_oneof = _SERVERREQUEST.oneofs_by_name['server_request_type']
_SERVERREQUEST.oneofs_by_name['server_request_type'].fields.append(
  _SERVERREQUEST.fields_by_name['inform_detach'])
_SERVERREQUEST.fields_by_name['inform_detach'].containing_oneof = _SERVERREQUEST.oneofs_by_name['server_request_type']
_SERVERREQUEST.oneofs_by_name['server_request_type'].fields.append(
  _SERVERREQUEST.fields_by_name['inform_teardown'])
_SERVERREQUEST.fields_by_name['inform_teardown'].containing_oneof = _SERVERREQUEST.oneofs_by_name['server_request_type']
_SERVERREQUEST.oneofs_by_name['server_request_type'].fields.append(
  _SERVERREQUEST.fields_by_name['inform_start'])
_SERVERREQUEST.fields_by_name['inform_start'].containing_oneof = _SERVERREQUEST.oneofs_by_name['server_request_type']
_SERVERRESPONSE.fields_by_name['result_communicate'].message_type = wandb_dot_proto_dot_wandb__internal__pb2._RESULT
_SERVERRESPONSE.fields_by_name['inform_init_response'].message_type = _SERVERINFORMINITRESPONSE
_SERVERRESPONSE.fields_by_name['inform_finish_response'].message_type = _SERVERINFORMFINISHRESPONSE
_SERVERRESPONSE.fields_by_name['inform_attach_response'].message_type = _SERVERINFORMATTACHRESPONSE
_SERVERRESPONSE.fields_by_name['inform_detach_response'].message_type = _SERVERINFORMDETACHRESPONSE
_SERVERRESPONSE.fields_by_name['inform_teardown_response'].message_type = _SERVERINFORMTEARDOWNRESPONSE
_SERVERRESPONSE.fields_by_name['inform_start_response'].message_type = _SERVERINFORMSTARTRESPONSE
_SERVERRESPONSE.oneofs_by_name['server_response_type'].fields.append(
  _SERVERRESPONSE.fields_by_name['result_communicate'])
_SERVERRESPONSE.fields_by_name['result_communicate'].containing_oneof = _SERVERRESPONSE.oneofs_by_name['server_response_type']
_SERVERRESPONSE.oneofs_by_name['server_response_type'].fields.append(
  _SERVERRESPONSE.fields_by_name['inform_init_response'])
_SERVERRESPONSE.fields_by_name['inform_init_response'].containing_oneof = _SERVERRESPONSE.oneofs_by_name['server_response_type']
_SERVERRESPONSE.oneofs_by_name['server_response_type'].fields.append(
  _SERVERRESPONSE.fields_by_name['inform_finish_response'])
_SERVERRESPONSE.fields_by_name['inform_finish_response'].containing_oneof = _SERVERRESPONSE.oneofs_by_name['server_response_type']
_SERVERRESPONSE.oneofs_by_name['server_response_type'].fields.append(
  _SERVERRESPONSE.fields_by_name['inform_attach_response'])
_SERVERRESPONSE.fields_by_name['inform_attach_response'].containing_oneof = _SERVERRESPONSE.oneofs_by_name['server_response_type']
_SERVERRESPONSE.oneofs_by_name['server_response_type'].fields.append(
  _SERVERRESPONSE.fields_by_name['inform_detach_response'])
_SERVERRESPONSE.fields_by_name['inform_detach_response'].containing_oneof = _SERVERRESPONSE.oneofs_by_name['server_response_type']
_SERVERRESPONSE.oneofs_by_name['server_response_type'].fields.append(
  _SERVERRESPONSE.fields_by_name['inform_teardown_response'])
_SERVERRESPONSE.fields_by_name['inform_teardown_response'].containing_oneof = _SERVERRESPONSE.oneofs_by_name['server_response_type']
_SERVERRESPONSE.oneofs_by_name['server_response_type'].fields.append(
  _SERVERRESPONSE.fields_by_name['inform_start_response'])
_SERVERRESPONSE.fields_by_name['inform_start_response'].containing_oneof = _SERVERRESPONSE.oneofs_by_name['server_response_type']
DESCRIPTOR.message_types_by_name['ServerShutdownRequest'] = _SERVERSHUTDOWNREQUEST
DESCRIPTOR.message_types_by_name['ServerShutdownResponse'] = _SERVERSHUTDOWNRESPONSE
DESCRIPTOR.message_types_by_name['ServerStatusRequest'] = _SERVERSTATUSREQUEST
DESCRIPTOR.message_types_by_name['ServerStatusResponse'] = _SERVERSTATUSRESPONSE
DESCRIPTOR.message_types_by_name['StringTupleValue'] = _STRINGTUPLEVALUE
DESCRIPTOR.message_types_by_name['SettingsValue'] = _SETTINGSVALUE
DESCRIPTOR.message_types_by_name['ServerInformInitRequest'] = _SERVERINFORMINITREQUEST
DESCRIPTOR.message_types_by_name['ServerInformInitResponse'] = _SERVERINFORMINITRESPONSE
DESCRIPTOR.message_types_by_name['ServerInformStartRequest'] = _SERVERINFORMSTARTREQUEST
DESCRIPTOR.message_types_by_name['ServerInformStartResponse'] = _SERVERINFORMSTARTRESPONSE
DESCRIPTOR.message_types_by_name['ServerInformFinishRequest'] = _SERVERINFORMFINISHREQUEST
DESCRIPTOR.message_types_by_name['ServerInformFinishResponse'] = _SERVERINFORMFINISHRESPONSE
DESCRIPTOR.message_types_by_name['ServerInformAttachRequest'] = _SERVERINFORMATTACHREQUEST
DESCRIPTOR.message_types_by_name['ServerInformAttachResponse'] = _SERVERINFORMATTACHRESPONSE
DESCRIPTOR.message_types_by_name['ServerInformDetachRequest'] = _SERVERINFORMDETACHREQUEST
DESCRIPTOR.message_types_by_name['ServerInformDetachResponse'] = _SERVERINFORMDETACHRESPONSE
DESCRIPTOR.message_types_by_name['ServerInformTeardownRequest'] = _SERVERINFORMTEARDOWNREQUEST
DESCRIPTOR.message_types_by_name['ServerInformTeardownResponse'] = _SERVERINFORMTEARDOWNRESPONSE
DESCRIPTOR.message_types_by_name['ServerRequest'] = _SERVERREQUEST
DESCRIPTOR.message_types_by_name['ServerResponse'] = _SERVERRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ServerShutdownRequest = _reflection.GeneratedProtocolMessageType('ServerShutdownRequest', (_message.Message,), {
  'DESCRIPTOR' : _SERVERSHUTDOWNREQUEST,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerShutdownRequest)
  })
_sym_db.RegisterMessage(ServerShutdownRequest)

ServerShutdownResponse = _reflection.GeneratedProtocolMessageType('ServerShutdownResponse', (_message.Message,), {
  'DESCRIPTOR' : _SERVERSHUTDOWNRESPONSE,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerShutdownResponse)
  })
_sym_db.RegisterMessage(ServerShutdownResponse)

ServerStatusRequest = _reflection.GeneratedProtocolMessageType('ServerStatusRequest', (_message.Message,), {
  'DESCRIPTOR' : _SERVERSTATUSREQUEST,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerStatusRequest)
  })
_sym_db.RegisterMessage(ServerStatusRequest)

ServerStatusResponse = _reflection.GeneratedProtocolMessageType('ServerStatusResponse', (_message.Message,), {
  'DESCRIPTOR' : _SERVERSTATUSRESPONSE,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerStatusResponse)
  })
_sym_db.RegisterMessage(ServerStatusResponse)

StringTupleValue = _reflection.GeneratedProtocolMessageType('StringTupleValue', (_message.Message,), {
  'DESCRIPTOR' : _STRINGTUPLEVALUE,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.StringTupleValue)
  })
_sym_db.RegisterMessage(StringTupleValue)

SettingsValue = _reflection.GeneratedProtocolMessageType('SettingsValue', (_message.Message,), {
  'DESCRIPTOR' : _SETTINGSVALUE,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.SettingsValue)
  })
_sym_db.RegisterMessage(SettingsValue)

ServerInformInitRequest = _reflection.GeneratedProtocolMessageType('ServerInformInitRequest', (_message.Message,), {

  'SettingsMapEntry' : _reflection.GeneratedProtocolMessageType('SettingsMapEntry', (_message.Message,), {
    'DESCRIPTOR' : _SERVERINFORMINITREQUEST_SETTINGSMAPENTRY,
    '__module__' : 'wandb.proto.wandb_server_pb2'
    # @@protoc_insertion_point(class_scope:wandb_internal.ServerInformInitRequest.SettingsMapEntry)
    })
  ,
  'DESCRIPTOR' : _SERVERINFORMINITREQUEST,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerInformInitRequest)
  })
_sym_db.RegisterMessage(ServerInformInitRequest)
_sym_db.RegisterMessage(ServerInformInitRequest.SettingsMapEntry)

ServerInformInitResponse = _reflection.GeneratedProtocolMessageType('ServerInformInitResponse', (_message.Message,), {
  'DESCRIPTOR' : _SERVERINFORMINITRESPONSE,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerInformInitResponse)
  })
_sym_db.RegisterMessage(ServerInformInitResponse)

ServerInformStartRequest = _reflection.GeneratedProtocolMessageType('ServerInformStartRequest', (_message.Message,), {

  'SettingsMapEntry' : _reflection.GeneratedProtocolMessageType('SettingsMapEntry', (_message.Message,), {
    'DESCRIPTOR' : _SERVERINFORMSTARTREQUEST_SETTINGSMAPENTRY,
    '__module__' : 'wandb.proto.wandb_server_pb2'
    # @@protoc_insertion_point(class_scope:wandb_internal.ServerInformStartRequest.SettingsMapEntry)
    })
  ,
  'DESCRIPTOR' : _SERVERINFORMSTARTREQUEST,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerInformStartRequest)
  })
_sym_db.RegisterMessage(ServerInformStartRequest)
_sym_db.RegisterMessage(ServerInformStartRequest.SettingsMapEntry)

ServerInformStartResponse = _reflection.GeneratedProtocolMessageType('ServerInformStartResponse', (_message.Message,), {
  'DESCRIPTOR' : _SERVERINFORMSTARTRESPONSE,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerInformStartResponse)
  })
_sym_db.RegisterMessage(ServerInformStartResponse)

ServerInformFinishRequest = _reflection.GeneratedProtocolMessageType('ServerInformFinishRequest', (_message.Message,), {
  'DESCRIPTOR' : _SERVERINFORMFINISHREQUEST,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerInformFinishRequest)
  })
_sym_db.RegisterMessage(ServerInformFinishRequest)

ServerInformFinishResponse = _reflection.GeneratedProtocolMessageType('ServerInformFinishResponse', (_message.Message,), {
  'DESCRIPTOR' : _SERVERINFORMFINISHRESPONSE,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerInformFinishResponse)
  })
_sym_db.RegisterMessage(ServerInformFinishResponse)

ServerInformAttachRequest = _reflection.GeneratedProtocolMessageType('ServerInformAttachRequest', (_message.Message,), {
  'DESCRIPTOR' : _SERVERINFORMATTACHREQUEST,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerInformAttachRequest)
  })
_sym_db.RegisterMessage(ServerInformAttachRequest)

ServerInformAttachResponse = _reflection.GeneratedProtocolMessageType('ServerInformAttachResponse', (_message.Message,), {

  'SettingsMapEntry' : _reflection.GeneratedProtocolMessageType('SettingsMapEntry', (_message.Message,), {
    'DESCRIPTOR' : _SERVERINFORMATTACHRESPONSE_SETTINGSMAPENTRY,
    '__module__' : 'wandb.proto.wandb_server_pb2'
    # @@protoc_insertion_point(class_scope:wandb_internal.ServerInformAttachResponse.SettingsMapEntry)
    })
  ,
  'DESCRIPTOR' : _SERVERINFORMATTACHRESPONSE,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerInformAttachResponse)
  })
_sym_db.RegisterMessage(ServerInformAttachResponse)
_sym_db.RegisterMessage(ServerInformAttachResponse.SettingsMapEntry)

ServerInformDetachRequest = _reflection.GeneratedProtocolMessageType('ServerInformDetachRequest', (_message.Message,), {
  'DESCRIPTOR' : _SERVERINFORMDETACHREQUEST,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerInformDetachRequest)
  })
_sym_db.RegisterMessage(ServerInformDetachRequest)

ServerInformDetachResponse = _reflection.GeneratedProtocolMessageType('ServerInformDetachResponse', (_message.Message,), {
  'DESCRIPTOR' : _SERVERINFORMDETACHRESPONSE,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerInformDetachResponse)
  })
_sym_db.RegisterMessage(ServerInformDetachResponse)

ServerInformTeardownRequest = _reflection.GeneratedProtocolMessageType('ServerInformTeardownRequest', (_message.Message,), {
  'DESCRIPTOR' : _SERVERINFORMTEARDOWNREQUEST,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerInformTeardownRequest)
  })
_sym_db.RegisterMessage(ServerInformTeardownRequest)

ServerInformTeardownResponse = _reflection.GeneratedProtocolMessageType('ServerInformTeardownResponse', (_message.Message,), {
  'DESCRIPTOR' : _SERVERINFORMTEARDOWNRESPONSE,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerInformTeardownResponse)
  })
_sym_db.RegisterMessage(ServerInformTeardownResponse)

ServerRequest = _reflection.GeneratedProtocolMessageType('ServerRequest', (_message.Message,), {
  'DESCRIPTOR' : _SERVERREQUEST,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerRequest)
  })
_sym_db.RegisterMessage(ServerRequest)

ServerResponse = _reflection.GeneratedProtocolMessageType('ServerResponse', (_message.Message,), {
  'DESCRIPTOR' : _SERVERRESPONSE,
  '__module__' : 'wandb.proto.wandb_server_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal.ServerResponse)
  })
_sym_db.RegisterMessage(ServerResponse)


_SERVERINFORMINITREQUEST_SETTINGSMAPENTRY._options = None
_SERVERINFORMSTARTREQUEST_SETTINGSMAPENTRY._options = None
_SERVERINFORMATTACHRESPONSE_SETTINGSMAPENTRY._options = None

_INTERNALSERVICE = _descriptor.ServiceDescriptor(
  name='InternalService',
  full_name='wandb_internal.InternalService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=2908,
  serialized_end=5936,
  methods=[
  _descriptor.MethodDescriptor(
    name='RunUpdate',
    full_name='wandb_internal.InternalService.RunUpdate',
    index=0,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._RUNRECORD,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._RUNUPDATERESULT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Attach',
    full_name='wandb_internal.InternalService.Attach',
    index=1,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._ATTACHREQUEST,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._ATTACHRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='TBSend',
    full_name='wandb_internal.InternalService.TBSend',
    index=2,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._TBRECORD,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._TBRESULT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RunStart',
    full_name='wandb_internal.InternalService.RunStart',
    index=3,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._RUNSTARTREQUEST,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._RUNSTARTRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetSummary',
    full_name='wandb_internal.InternalService.GetSummary',
    index=4,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._GETSUMMARYREQUEST,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._GETSUMMARYRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SampledHistory',
    full_name='wandb_internal.InternalService.SampledHistory',
    index=5,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._SAMPLEDHISTORYREQUEST,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._SAMPLEDHISTORYRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='PollExit',
    full_name='wandb_internal.InternalService.PollExit',
    index=6,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._POLLEXITREQUEST,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._POLLEXITRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Shutdown',
    full_name='wandb_internal.InternalService.Shutdown',
    index=7,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._SHUTDOWNREQUEST,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._SHUTDOWNRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RunExit',
    full_name='wandb_internal.InternalService.RunExit',
    index=8,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._RUNEXITRECORD,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._RUNEXITRESULT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RunPreempting',
    full_name='wandb_internal.InternalService.RunPreempting',
    index=9,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._RUNPREEMPTINGRECORD,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._RUNPREEMPTINGRESULT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Metric',
    full_name='wandb_internal.InternalService.Metric',
    index=10,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._METRICRECORD,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._METRICRESULT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='PartialLog',
    full_name='wandb_internal.InternalService.PartialLog',
    index=11,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._PARTIALHISTORYREQUEST,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._PARTIALHISTORYRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Log',
    full_name='wandb_internal.InternalService.Log',
    index=12,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._HISTORYRECORD,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._HISTORYRESULT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Summary',
    full_name='wandb_internal.InternalService.Summary',
    index=13,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._SUMMARYRECORD,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._SUMMARYRESULT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Config',
    full_name='wandb_internal.InternalService.Config',
    index=14,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._CONFIGRECORD,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._CONFIGRESULT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Files',
    full_name='wandb_internal.InternalService.Files',
    index=15,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._FILESRECORD,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._FILESRESULT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Output',
    full_name='wandb_internal.InternalService.Output',
    index=16,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._OUTPUTRECORD,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._OUTPUTRESULT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Telemetry',
    full_name='wandb_internal.InternalService.Telemetry',
    index=17,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__telemetry__pb2._TELEMETRYRECORD,
    output_type=wandb_dot_proto_dot_wandb__telemetry__pb2._TELEMETRYRESULT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Alert',
    full_name='wandb_internal.InternalService.Alert',
    index=18,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._ALERTRECORD,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._ALERTRESULT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Artifact',
    full_name='wandb_internal.InternalService.Artifact',
    index=19,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._ARTIFACTRECORD,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._ARTIFACTRESULT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LinkArtifact',
    full_name='wandb_internal.InternalService.LinkArtifact',
    index=20,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._LINKARTIFACTRECORD,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._LINKARTIFACTRESULT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ArtifactSend',
    full_name='wandb_internal.InternalService.ArtifactSend',
    index=21,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._ARTIFACTSENDREQUEST,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._ARTIFACTSENDRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ArtifactPoll',
    full_name='wandb_internal.InternalService.ArtifactPoll',
    index=22,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._ARTIFACTPOLLREQUEST,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._ARTIFACTPOLLRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CheckVersion',
    full_name='wandb_internal.InternalService.CheckVersion',
    index=23,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._CHECKVERSIONREQUEST,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._CHECKVERSIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Pause',
    full_name='wandb_internal.InternalService.Pause',
    index=24,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._PAUSEREQUEST,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._PAUSERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Resume',
    full_name='wandb_internal.InternalService.Resume',
    index=25,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._RESUMEREQUEST,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._RESUMERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Status',
    full_name='wandb_internal.InternalService.Status',
    index=26,
    containing_service=None,
    input_type=wandb_dot_proto_dot_wandb__internal__pb2._STATUSREQUEST,
    output_type=wandb_dot_proto_dot_wandb__internal__pb2._STATUSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ServerShutdown',
    full_name='wandb_internal.InternalService.ServerShutdown',
    index=27,
    containing_service=None,
    input_type=_SERVERSHUTDOWNREQUEST,
    output_type=_SERVERSHUTDOWNRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ServerStatus',
    full_name='wandb_internal.InternalService.ServerStatus',
    index=28,
    containing_service=None,
    input_type=_SERVERSTATUSREQUEST,
    output_type=_SERVERSTATUSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ServerInformInit',
    full_name='wandb_internal.InternalService.ServerInformInit',
    index=29,
    containing_service=None,
    input_type=_SERVERINFORMINITREQUEST,
    output_type=_SERVERINFORMINITRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ServerInformStart',
    full_name='wandb_internal.InternalService.ServerInformStart',
    index=30,
    containing_service=None,
    input_type=_SERVERINFORMSTARTREQUEST,
    output_type=_SERVERINFORMSTARTRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ServerInformFinish',
    full_name='wandb_internal.InternalService.ServerInformFinish',
    index=31,
    containing_service=None,
    input_type=_SERVERINFORMFINISHREQUEST,
    output_type=_SERVERINFORMFINISHRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ServerInformAttach',
    full_name='wandb_internal.InternalService.ServerInformAttach',
    index=32,
    containing_service=None,
    input_type=_SERVERINFORMATTACHREQUEST,
    output_type=_SERVERINFORMATTACHRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ServerInformDetach',
    full_name='wandb_internal.InternalService.ServerInformDetach',
    index=33,
    containing_service=None,
    input_type=_SERVERINFORMDETACHREQUEST,
    output_type=_SERVERINFORMDETACHRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ServerInformTeardown',
    full_name='wandb_internal.InternalService.ServerInformTeardown',
    index=34,
    containing_service=None,
    input_type=_SERVERINFORMTEARDOWNREQUEST,
    output_type=_SERVERINFORMTEARDOWNRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_INTERNALSERVICE)

DESCRIPTOR.services_by_name['InternalService'] = _INTERNALSERVICE

# @@protoc_insertion_point(module_scope)
