# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: wandb/proto/wandb_base.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='wandb/proto/wandb_base.proto',
  package='wandb_internal',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1cwandb/proto/wandb_base.proto\x12\x0ewandb_internal\"6\n\x0b_RecordInfo\x12\x11\n\tstream_id\x18\x01 \x01(\t\x12\x14\n\x0c_tracelog_id\x18\x64 \x01(\t\"!\n\x0c_RequestInfo\x12\x11\n\tstream_id\x18\x01 \x01(\t\"#\n\x0b_ResultInfo\x12\x14\n\x0c_tracelog_id\x18\x64 \x01(\tb\x06proto3'
)




__RECORDINFO = _descriptor.Descriptor(
  name='_RecordInfo',
  full_name='wandb_internal._RecordInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='stream_id', full_name='wandb_internal._RecordInfo.stream_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='_tracelog_id', full_name='wandb_internal._RecordInfo._tracelog_id', index=1,
      number=100, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=48,
  serialized_end=102,
)


__REQUESTINFO = _descriptor.Descriptor(
  name='_RequestInfo',
  full_name='wandb_internal._RequestInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='stream_id', full_name='wandb_internal._RequestInfo.stream_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=104,
  serialized_end=137,
)


__RESULTINFO = _descriptor.Descriptor(
  name='_ResultInfo',
  full_name='wandb_internal._ResultInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='_tracelog_id', full_name='wandb_internal._ResultInfo._tracelog_id', index=0,
      number=100, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=139,
  serialized_end=174,
)

DESCRIPTOR.message_types_by_name['_RecordInfo'] = __RECORDINFO
DESCRIPTOR.message_types_by_name['_RequestInfo'] = __REQUESTINFO
DESCRIPTOR.message_types_by_name['_ResultInfo'] = __RESULTINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_RecordInfo = _reflection.GeneratedProtocolMessageType('_RecordInfo', (_message.Message,), {
  'DESCRIPTOR' : __RECORDINFO,
  '__module__' : 'wandb.proto.wandb_base_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal._RecordInfo)
  })
_sym_db.RegisterMessage(_RecordInfo)

_RequestInfo = _reflection.GeneratedProtocolMessageType('_RequestInfo', (_message.Message,), {
  'DESCRIPTOR' : __REQUESTINFO,
  '__module__' : 'wandb.proto.wandb_base_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal._RequestInfo)
  })
_sym_db.RegisterMessage(_RequestInfo)

_ResultInfo = _reflection.GeneratedProtocolMessageType('_ResultInfo', (_message.Message,), {
  'DESCRIPTOR' : __RESULTINFO,
  '__module__' : 'wandb.proto.wandb_base_pb2'
  # @@protoc_insertion_point(class_scope:wandb_internal._ResultInfo)
  })
_sym_db.RegisterMessage(_ResultInfo)


# @@protoc_insertion_point(module_scope)
